# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import uuid
import time
import json

from threading import Event
from typing import Optional
from datetime import datetime
from socketio import Client
from neon_utils.logger import LOG

from klat_connector.utils.legacy_socket_adapter import SocketIOCompat
from klat_connector.utils.encryptor import generate_hash
from klat_connector.exceptions import *
from klat_connector.api.klat_api_abc import KlatApiABC
from klat_connector.utils.html_utils import clean_html_entities


FIRST_SHOUT = "New API Conversation."


class KlatApi(KlatApiABC):
 
    def __init__(self, socket, dom="klattalk.com", init_nick=None,
                 start_conenction: bool = True):
        """
        Initializes a Klat connection to the passed socket. Initially connects as a guest user in the passed domain.
        :param socket: socketio.client Client socket connection to the Klat server
        :param dom: initial Klat domain to connect to
        """
        self._uid = str(uuid.uuid1())
        self._users = None
        self._nick = init_nick or ""
        self._title = None
        self._cid = None
        self._dom = dom
        self._log = None
        self._profile = None  # Populated on "nickname" or "mobile login return"
        self._devices = None  # Populated on "nickname"
        self._login = 0       # 0=in-progress 1=guest 2=logged in

        self._users_last_updated = 0
        self._users_list_cache = 2
        self._socket_id = None
        self._client = "api"
        self._get_users_timeout = 5

        # self._await_log = False
        self._await_log = Event()
        self._await_users_list = Event()
        self._await_change_conversation = Event()
        self._await_new_conversation = Event()
        self._await_login = Event()
        self._await_logout = Event()
        self._created_cid = None
        self._created_title = None
        self._created_needs_shout = False

        self.instance = str(round(time.time()))
        self._socket: Client = socket
        self.klat_ready = Event()
        self.klat_ready.clear()

        if start_conenction:
            self.run()

    def run(self, *args, **kwargs):
        self._setup_listeners()

        if not self._socket.connected:
            raise ConnectionError(f"{self._socket.sid} @ "
                                  f"{self._socket.connection_url}")
        self._start_connection()

        if not self.klat_ready.wait(30):
            LOG.error(f"Timed out waiting for Klat Ready!")
            self._socket.disconnect()
            raise KlatAPIError("Klat server connection not established!")

    @property
    def socket(self) -> Client:
        if not self._socket.connected:
            LOG.error("Socket disconnected")
            if self._socket.eio.state != "disconnected":
                LOG.error("Socket Partially disconnected")
                self._socket.eio.disconnect()
        return self._socket

    @property
    def ready(self) -> bool:
        """
        Added for backwards-compat. Use klat_ready.wait() instead of ready
        :return:
        """
        LOG.warning(f"This method is deprecated. Use `self.klat_ready` event.")
        return self.klat_ready.is_set()

    @property
    def connected(self) -> bool:
        """
        Return the connection state of the SocketIO object
        :returns: True if SocketIO connected
        """
        return self.socket.connected

    @property
    def chat_log(self) -> Optional[list]:
        """
        Return a list chat log if initialized
        :returns: chat log from server if initialized, else None
        """
        return self._log

    @property
    def nick(self) -> str:
        """
        Return the server-specified nick associated with this connection
        :returns: requested nick or actual nick returned by server after init
        """
        return self._nick

    @property
    def title(self) -> Optional[str]:
        """
        Return the title of the current conversation
        :returns: title last specified by server if initialized, else None
        """
        return self._title

    @property
    def logged_in(self) -> int:
        """
        Wraps self._login to return login status
        :return: 0 = Disconnected, 1 = Guest, 2 = User
        """
        return self._login

    @property
    def conversation_is_proctored(self) -> bool:
        """
        Determines if there is a Proctor in the current conversation
        :return: Boolean if there is a proctor in the current conversation
        """
        users = [u.lower() for u in self.conversation_users]
        if users and len(users) > 0:
            LOG.debug(users)
            return "proctor" in users
        else:
            LOG.error(f"Error updating conversation users."
                      f"Proctored={self._dom == 'chatbotsforum.org'}")
            return self._dom == "chatbotsforum.org"

    @property
    def conversation_users(self) -> list:
        """
        Returns an updated list of users in the current conversation
        :return: list of nicks in the current conversation
        """
        if time.time() - self._users_last_updated > self._users_list_cache or not self._users:
            # 'get nicks for cid', cid
            self.socket.emit("get nicks for cid", self._cid)
            self._await_users_list.clear()
            if not self._await_users_list.wait(self._get_users_timeout):
                LOG.error(f'Missing response to users: {self._cid}, fallback to old value')

        if self._users != list(dict.fromkeys(self._users)):
            LOG.warning(f"Duplicates found in users: {self._users}")
            self._users = list(dict.fromkeys(self._users))
        return self._users

    def is_current_cid(self, cid: str) -> bool:
        """
        Checks if the passed cid is the current cid.
        :param cid: cid to check
        :return: True if cid is the current cid
        """
        return cid == self._cid

    # Exposed API Methods
    def send_shout(self, shout: str, cid: str = None, dom: str = None, **_):
        """
        Shout into the current conversation or else passed dom/cid
        :param shout: text to shout
        :param cid: CID to send shout into (Private cid for @user shout)
        :param dom: Domain associated with cid
        """
        cookies = json.loads('{"ShareMyRecordings": 1}')
        dest_dom = dom or self._dom
        dest_cid = cid or self._cid

        try:
            self._send_shout(shout, dest_dom, dest_cid, cid, cookies)
        except ConnectionError:
            LOG.error(f"Socket disconnected skipping sending: {self.nick} - {shout}")
        except Exception as e:
            LOG.error(e)

    def search_shout(self, search: str, shout_context: int = 10, shout_time: int = 30):
        """
        Search for a shout with the given input
        :param search: input string (regex supported) to match
        :param shout_context: Number of contextual shouts to return (before or after)
        :param shout_time: Max seconds of context to provide (before or after)
        """
        # (search, num_context, max_time)
        self.socket.emit("search shouts by phrase", (search, shout_context, 1000*shout_time))

    def login_klat(self, user: str, password: str):
        """
        Login to klat.com with the passed username and password
        :param user: klat nick
        :param password: cleartext login password
        """

        if not self.klat_ready.wait(30):
            LOG.error(f"Timed out waiting for Klat Ready!")
            self.socket.disconnect()
            raise KlatAPIError("Klat server connection not established!")
        self._login = 0
        self._await_login.clear()
        hash_pass = generate_hash(password)
        self.socket.emit("check login", (self._nick, user, self._dom, hash_pass, None, self._cid, self._uid))
        if not self._await_login.wait(30):
            self.socket.disconnect()
            raise LoginError("Error in response to 'check login'")

    def logout_klat(self):
        """
        Log out of current account
        """
        # (nick, dom, color, cid, kinstance, col, page name)
        self.socket.emit("logout", (self._nick, self._dom, None, self._cid, self.instance, "col1", "1col"))
        self._await_logout.clear()
        if not self._await_logout.wait(30):
            self.socket.disconnect()
            raise LoginError("Error in response to 'logout'")

    def register_klat(self, username, password):
        """
        Register a new user
        """
        self._login = 0
        self._await_login.clear()
        self.socket.emit("save login", (username, self._dom, generate_hash(password), "rgb(0, 0, 0)", self._cid,
                                        str(self.instance), "col1", "1col", "1", "1", 1.0, "Google US English",
                                        "Joanna", "female", "en-us", "en-us", "en", 12, "imperial", "MDY", "Renton",
                                        "Washington", "United States", self._uid, None))
        if not self._await_login.wait(30):
            self.socket.disconnect()
            raise LoginError("Error in response to 'save login'")

    def change_domain(self, new_dom: str):
        """
        Change to a new conversation domain
        :param new_dom: domain to go to
        """
        # (whichWay, currentCid, domainToGoTo, doChangeDomain)
        if new_dom.lower() == "private":
            new_dom = "Private"
        else:
            new_dom = new_dom.lower()
        # TODO: verify domain is valid DM
        self._await_change_conversation.clear()
        self.socket.emit("change domain", ("this one", self._cid, new_dom))
        if not self._await_change_conversation.wait(30):
            self.socket.disconnect()
            raise ChangeConversationError(f"Change Domain Error! new_dom={new_dom}")
        if new_dom == "Private" and self._nick.lower() not in self._title.lower():
            self.socket.disconnect()
            raise ChangeConversationError(f"Invalid Private conversation joined! nick={self._nick}|title={self.title}")

    def change_conversation(self, new_cid: str, new_dom: str):
        """
        Change to a new conversation by cid
        :param new_cid: cid to go to
        :param new_dom: domain for new_cid
        """
        self._await_change_conversation.clear()
        # (whichWay, currentCid, domainToGoTo, doChangeDomain)
        self.socket.emit("change domain", ("conversation only", new_cid, new_dom))

        if not self._await_change_conversation.wait(30):
            self.socket.disconnect()
            raise ChangeConversationError(f"Change Conversation Error! new_dom={new_dom}|new_cid={new_cid}")

    def get_private_conversation(self, users: list) -> Optional[str]:
        """
        Helper function that wraps create_conversation and builds a title from a list of desired users.
        Private conversation is created if not exists and cid is returned.
        :param users: list of users to include in the private conversation
        :return: cid of created conversation
        """
        if self._nick not in users:
            LOG.info(f"adding creating user to conversation: {self._nick}")
            users.append(self._nick)
        title = f"!PRIVATE:{','.join(users)}"
        new_cid = self.create_conversation("Private", title)
        return new_cid

    def create_conversation(self, dom: str, title: str, goto: bool = False,
                            first_shout: str = FIRST_SHOUT) -> Optional[str]:
        """
        Creates a new Klat conversation with the given parameters
        :param dom: Domain for new conversation
        :param title: Title of new conversation
        :param goto: Flag switch to created conversation
        :param first_shout: First shout to send into new conversation
        :return: Created cid or null if failed
        """
        self._created_cid: Optional[str] = None
        self._await_new_conversation.clear()

        if title.startswith("!PRIVATE:") and dom != "Private":
            LOG.error(f"Attempt to create private conversation in dom={dom}")
            dom = "Private"

        # (domain, nick, cid, title, ImageURL, ArticleURL, msgFromNeon, PassHash, FirstShout, keychain)
        self.socket.emit("create conversation", (dom, self._nick, self._cid, title, "", "", None, "",
                                                 first_shout, None))

        if not self._await_new_conversation.wait(30):
            self.socket.disconnect()
            raise CreateConversationError(f"New conversation timed out! dom={dom}, title={title}")
        if self._created_cid:
            if self._created_needs_shout:
                self.send_shout(first_shout, self._created_cid, dom)
            if goto:
                self._title = self._created_title
                self.change_conversation(self._created_cid, dom)
        else:
            # TODO: Should this raise an exception? DM
            LOG.error(f"Error while creating {title} on {dom}")
        return self._created_cid

    def get_conversation_log(self, follows: int = 10, context: int = 20,
                             shouts: int = 50, age: int = 60000, last: int = 0):
        """
        Requests an update of the conversation history for the current conversation
        :param follows: Max number of followed conversations to get
        :param context: Max number of context shouts to get surrounding crossposts
        :param shouts: Max number of shouts to get
        :param age: Max age of shouts to get in seconds
        :param last: Millis time of oldest shout to get
        """
        self._log = None

        max_follows = str(follows)
        max_context = str(context)
        max_shouts = str(shouts)
        max_time = str(age)
        last_shout = last
        # self._await_log = True
        self._await_log.clear()
        # (myDomain, oldConversation, myConversation, myNick, myFollowsKey, myColumn, myMaxFollowedConversations,
        # myContextMax, myContextTime, nano, lastShout, numShouts, function(err){})
        self.socket.emit("get log", (self._dom, "", self._cid, self._nick, "", "col1",
                                     max_follows, max_context, max_time, self._client, last_shout, max_shouts))

        if not self._await_log.wait(30):
            self.socket.disconnect()
            raise KlatAPIError("No conversation log received from server!")

    # Handlers to be overridden
    def handle_incoming_shout(self, user: str, shout: str, cid: str, dom: str, timestamp: str):
        """
        This function should be overridden to handle incoming messages
        :param user: user associated with shout
        :param shout: user's shout
        :param cid: cid shout belongs to
        :param dom: domain conversation belongs to
        :param timestamp: time of incoming shout
        """
        pass

    def handle_login_return(self, status: int):
        """
        This function should be overridden to handle login return status
        :param status: non-success return code
        """
        pass

    def handle_search_results(self, kind: str, results: dict):
        """
        This function should be overridden to handle search results
        :param kind: elements searched (i.e. "shouts")
        :param results: dict of matched and contextual shouts by cid
        """
        pass

    def handle_socket_reconnected(self):
        """
        This function should be overridden to handle a server socket reconnection
        """
        pass

    # Socket Listeners
    def _on_connect(self):
        """
        Handler for socket connection
        """
        LOG.info("Chat Server Connected")
        if self._login != 0:
            # Handle reconnections here
            self.handle_socket_reconnected()
            self._start_connection()

    @staticmethod
    def _on_disconnect():
        """
        Handler for socket disconnection
        """
        # self.connected = False
        LOG.warning("Chat Server Socket Disconnected!")

    def _on_connect_error(self, data):
        """
        Handler for socket connection error
        """
        LOG.error(f"SocketIO Error: {data}")

    @staticmethod
    def _on_reconnect():
        """
        Handler for socket reconnection
        """
        # self.connected = True
        LOG.warning("SocketIO Reconnected")

    def _on_conversation_list(self, *args):
        """
        Handler for "conversation list". Emits "nickname"
        :param args: Socket Arguments
        """
        self._socket_id = args[2]
        conversations = args[0][0]  # JsonArray of JsonObjects
        new_conversation = self._parse_conversations_list(conversations)
        c_nick = self._nick or new_conversation["cNick"]
        l_uname = self._nick or new_conversation["lUname"]
        new_title = new_conversation["title"]
        # (myCNick, luname, ldom, myConversation, userTitle, myFollowsKey, myColumn, myKinstance, myPageName,
        # myTtsLanguage, mySecondTtsLanguage, mySttLanguage, nano, uid, function...
        self.socket.emit("nickname",
                         (c_nick, l_uname, self._dom, self._cid, new_title, "", "col1", float(self.instance),
                          "1col", "en-us", "en-us", "en", self._client, self._uid, None))

    def _on_nickname(self, *args):
        """
        Handler for "nickname". After this, _profile is populated and we can send/receive messages
        :param args: Socket Arguments
        :return:
        """
        self._nick = args[0]
        self._devices = args[1]
        self._profile = args[2]
        LOG.info(f"Logged in as {self._nick}")
        self._login = 1
        self._await_logout.set()
        self.klat_ready.set()

    def _on_user_message(self, *args):
        """
        Handler for "user message" (incoming shouts)
        :param args: Socket Arguments
        """
        LOG.debug(args)
        shout = clean_html_entities(args[0])
        user = args[1].split('#', 1)[0]
        s_cid = str(args[2])
        s_dom = args[5]
        s_time = round(int(args[3]) / 1000)
        p_time = datetime.fromtimestamp(s_time).strftime("%I:%M:%S %p")
        self.handle_incoming_shout(user, shout, s_cid, s_dom, p_time)

    def _on_login_return(self, *args):
        """
        Handler for "mobile login return". This is where nick is updated
        :param args: Socket Arguments
        """
        self._users = None  # Clear users to force checking with server
        LOG.debug(len(args))
        login_status = args[0]
        if login_status == 0:
            # Successful Login
            self._nick = args[1]
            if not self._profile:
                LOG.error("Profile not defined at login return!")
                self._profile = {}
            self._profile = {"display_nick": args[1],
                             "color": "",
                             "avatar_url": args[4],
                             "pass": "",
                             "mail": args[3],
                             "login": time.time(),
                             "timezone": None,
                             "ip": self._profile.get("ip", "0.0.0.0"),
                             "speech_rate": self._profile.get("speech_rate", "1.0"),
                             "speech_pitch": self._profile.get("speech_pitch", "1.0"),
                             "volume": float(args[33] or 0.0),
                             "speech_voice": args[11],
                             "ai_speech_voice": args[12],
                             "time_format": args[17],
                             "date_format": args[18],
                             "unit_measure": args[19],
                             "stt_language": args[16],
                             "tts_language": args[14],
                             "tts_secondary_language": args[15],
                             "tts_voice_gender": self._profile.get("tts_voice_gender", "female")}

            # Goto new conversation
            if args[27]:
                LOG.info("Create new Private!")
            elif self._dom == "Private":
                new_cid = str(args[26])
                self._cid = new_cid
            self._login = 2
        elif login_status == 777:
            self.logout_klat()
            self._login = -1
            self._nick = ''
        elif login_status == "mock":
            self._profile = {}
            self._nick = args[1]
            self._login = 2  # Logged in on mock server
            login_status = 0
        else:
            self._login = 1
            self._await_logout.set()
        LOG.info(f"Logged in as: {self._nick}")
        self._await_login.set()
        self.handle_login_return(login_status)

    def _on_new_domain(self, *args):
        """
        Handler for "new domain". This is where dom and cid are updated
        :param args: Socket Arguments
        """
        LOG.debug(args)
        self._cid = str(args[0])
        self._users = args[1] or []
        if self._nick not in self._users:
            LOG.warning("Nick not included in new conversation!")
            self._users.append(self._nick)

        self._users_last_updated = time.time()
        data = args[2]
        self._dom = data.get("dom")
        self._title = data.get("title")
        LOG.info(f">>>Got new cid: {self._cid} with title: {self._title}")
        self._await_change_conversation.set()

    def _on_chat_log(self, *args):
        """
        Handler for "chat log" Populates shout history for the current conversation
        :param args: Socket Arguments
        """
        history = args[0]
        users = args[1]
        self._log = history or []
        self._users = users
        if self._nick not in users:
            LOG.warning("Nick not included in new conversation!")
            self._users.append(self._nick)
        # self._await_log = False
        self._await_log.set()

    def _on_log_out(self, *args):
        """
        Handler for "log out successful"
        :param args: Socket Arguments
        """
        self._users = None
        LOG.debug(args)
        self._login = 0
        self._nick = ''
        if self._dom == "Private":
            self._cid = None
        self._start_connection()

    def _on_search_shouts_by_phrase_return(self, *args):
        """
        Handler for "search shouts by phrase return"
        :param args: Socket Arguments
        """
        LOG.debug(args)
        results = dict()
        for key, val in dict(args[0]).items():
            new_key = key.split(":")[0]
            if new_key in results.keys():
                results[new_key].append(val)
            else:
                results[new_key] = [val]
        self.handle_search_results("shouts", results)

    def _on_nicks_list(self, *args):
        """
        Handler for "new nick list"
        :param args: Socket Arguments
        """
        # if self._users:
        #     LOG.warning(f"users not empty, this probably came late!")
        LOG.info(f"Got new users: {args[1]}")
        self._users_last_updated = time.time()
        self._users = args[1]
        self._await_users_list.set()

    def _on_save_login_return(self, *args):
        """
        Handler for "save login return"
        :param args: Socket Arguments
        """
        self._users = None  # Clear users to force checking with server
        register_status = args[2]
        if register_status == "Name already exists.":
            success = False
        elif register_status == "Name cannot contain non-alphanumeric characters or reserved words.":
            success = False
        elif register_status == "Save login successful.":
            success = True
        else:
            success = False

        confirmed_nick = args[1]
        if success:
            LOG.info(f"Logged in as {confirmed_nick}")
            self._nick = confirmed_nick
            if self._dom == "Private":
                self._cid = None
                self.change_domain("Private")
            self._login = 2
        else:
            self._login = 1
        self._await_login.set()

    def _on_create_conversation_return(self, *args):
        """
        Handler for "create conversation return
        :param args: Socket Arguments
        """
        ret_cid: str = str(args[0])
        need_shout = False
        if ret_cid.startswith("Title exists - "):
            new_cid = ret_cid.split("-", 1)[1].strip()
            new_title: str = args[1]
            LOG.info(f"Cid exists: {ret_cid}")

        elif ret_cid.startswith("Cannot start non-private"):
            LOG.error(ret_cid)
            new_cid = None
            new_title = ""
        else:
            new_cid = ret_cid
            new_dom: str = args[3]
            new_title: str = args[4]
            # hash_pass: str = args[6]
            LOG.info(f"Created cid: {new_cid} on {new_dom}")
            need_shout = True

        self._created_cid = new_cid
        self._created_title = new_title
        self._created_needs_shout = need_shout
        self._await_new_conversation.set()

    # Internal Functions
    def _parse_conversations_list(self, conversations_list: dict):
        """
        Called by _on_conversation_list to parse returned JSON object and collect new conversation data
        :param conversations_list: conversation data
        :return: current conversation data
        """
        dom_data = conversations_list.get("row")
        new_dom = dom_data.get("dom")
        new_title = dom_data.get("title")
        new_cid = str(dom_data.get("cid", conversations_list.get("cid", [""])[0]))
        self._cid = new_cid
        self._dom = new_dom
        self._title = new_title

        c_nick = self._nick or ""
        if new_dom == "Private":
            l_uname = "Private:User"
        else:
            l_uname = "klat:User"

        return {"cNick": c_nick,
                "lUname": l_uname,
                "title": new_title}

    def _send_shout(self, shout: str, dest_dom: str, dest_cid: str, orig_cid: str, cookies: json):
        """
        Internal function that sends a shout into the current conversation and throws an exception if not connected
        :param shout: text to shout
        :param dest_dom: Destination domain for the shout
        :param dest_cid: Destination cid for the shout
        :param orig_cid: CID the shout originated from
        :param cookies: JSON object associated with this instance
        """

        if not self.socket.connected:
            raise ConnectionError(f"Socket disconnected skipping sending: {self.nick} - {shout}")
        else:
            # (msg, dom, cid, img, video, mode, imgData, showAnswers, audioBlobs, audioFileName, sttLanguage,
            # ttsLanguage, ttsSecondLanguage, ttsGender, needsTranslation, cidForLink, cookies, name, fn)
            self.socket.emit("user message", (shout, dest_dom, dest_cid, "", "", self._client, "", 0, None, False,
                                              self._profile.get("stt_language", "en"),
                                              self._profile.get("tts_language", "en-US"),
                                              self._profile.get("tts_secondary_language"), "male",
                                              self._profile.get("stt_language", "en") != "en", orig_cid,
                                              cookies, False, None, False, False, False))

    def _start_connection(self):
        """
        Initializes a new connection to the Klat server
        """
        nick = self._nick
        cid = self._cid or ""
        domain = self._dom or "Private"
        welcome_message = ""
        # (pCid, cookieCid, mode, dom, nick, tab, hash, url, klatGuid, pageName, column, kinstance, allCookies,
        # nano, welcomeMessage, fn)
        self.socket.emit("get conversation list 2", (cid, "", "init", domain, nick, "1", "", "",
                                                     self.instance, "1col", "col1", self.instance, "", self._client,
                                                     welcome_message))

    def _stop_connection(self):
        LOG.info('Connection stopped')

    def _setup_listeners(self):
        """
        Starts all Klat event listeners
        """
        self.socket.on("connect", self._on_connect)
        self.socket.on("disconnect", self._on_disconnect)
        self.socket.on("connect_error", self._on_connect_error)
        self.socket.on("reconnect", self._on_reconnect)
        self.socket.on("conversation list", self._on_conversation_list)
        self.socket.on("nickname", self._on_nickname)
        self.socket.on("mobile user message", self._on_user_message)
        self.socket.on("mobile login return", self._on_login_return)
        self.socket.on("new domain", self._on_new_domain)
        self.socket.on("chat log", self._on_chat_log)
        self.socket.on("log out successful", self._on_log_out)
        self.socket.on("search shouts by phrase return", self._on_search_shouts_by_phrase_return)
        self.socket.on("new nick list", self._on_nicks_list)
        self.socket.on("save login return", self._on_save_login_return)
        self.socket.on("create conversation return", self._on_create_conversation_return)
        if isinstance(self.socket, SocketIOCompat):
            self.socket.run_forever()

    def shutdown(self):
        """
        Shutdown this connection; stop all consumers and close all connections
        """
        self._socket.disconnect()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import unittest
import pytest

from time import sleep
from socketio import Client

from klat_connector.utils.sio_connection_utils import start_socket, SocketIOCompat


class EncryptorTests(unittest.TestCase):
    def test_generate_hash(self):
        from klat_connector.utils.encryptor import generate_hash
        hashed = generate_hash("test123")
        self.assertIsInstance(hashed, list)
        self.assertTrue(all((isinstance(x, int) for x in hashed)))


class HtmlUtilsTests(unittest.TestCase):
    def test_clean_html_entities_valid(self):
        from klat_connector.utils.html_utils import clean_html_entities
        cleaned = clean_html_entities("&quot;hello&quot;")
        self.assertEqual(cleaned, '"hello"')
        cleaned = clean_html_entities("\"hello\"")
        self.assertEqual(cleaned, '"hello"')


class LegacySocketAdapterTests(unittest.TestCase):
    from klat_connector.utils.legacy_socket_adapter import SocketIOCompat
    # TODO


class MachServerTests(unittest.TestCase):
    from klat_connector.utils.mach_server import MachKlatServer
    from klat_connector.klat_api import KlatApi
    from klat_connector.utils.sio_connection_utils import start_socket
    sio_server = MachKlatServer()
    sleep(1)  # Pad to allow mach server to start up
    sio_client = KlatApi(start_socket("0.0.0.0"))

    @classmethod
    def tearDownClass(cls) -> None:
        cls.sio_server.shutdown_server()
        cls.sio_client.socket.disconnect()

    @pytest.mark.timeout(5)
    def test_mock_users(self):
        conversation_users = self.sio_client.conversation_users
        this_user = self.sio_client.nick
        self.assertEqual(conversation_users, [this_user])

    @pytest.mark.timeout(5)
    def test_mock_login(self):
        self.sio_client.login_klat("testlogin", "")
        self.assertEqual(self.sio_client.conversation_users, ["testlogin"])

    @pytest.mark.timeout(5)
    def test_mock_logout(self):
        self.assertFalse(self.sio_client.nick.startswith("guest"))
        self.sio_client.logout_klat()
        self.assertTrue(self.sio_client.nick.startswith("guest"))
        self.assertEqual(self.sio_client.conversation_users, [self.sio_client.nick])

    @pytest.mark.timeout(5)
    def test_mock_domain(self):
        self.assertEqual(self.sio_client._dom, "klattalk.com")
        klat_cid = self.sio_client._cid
        self.sio_client.change_domain("Private")
        self.assertFalse(self.sio_client.is_current_cid(klat_cid))
        self.sio_client.change_domain("klattalk.com")
        self.assertTrue(self.sio_client.is_current_cid(klat_cid))

    @pytest.mark.timeout(5)
    def test_mock_register(self):
        self.sio_client.register_klat("new_user", "password")
        self.assertEqual(self.sio_client.nick, "new_user")
        self.sio_client.logout_klat()
        self.assertTrue(self.sio_client.nick.startswith("guest"))

    def test_mock_login_to_domain(self):
        from klat_connector.klat_api import KlatApi
        api = KlatApi(start_socket("0.0.0.0"), "chatbotsforum")
        api.login_klat("test_user", "password")
        self.assertEqual(api.nick, "test_user")
        self.assertEqual(api._dom, "chatbotsforum")
        api.socket.disconnect()


class SocketConnectionTests(unittest.TestCase):
    def test_start_socket(self):
        socket = start_socket()
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.disconnect()

    def test_start_socket_bad_port_timeout(self):
        socket = start_socket("2222.us", 80, 3)
        self.assertFalse(socket.connected)

    @pytest.mark.skip("Server is decommissioned")
    def test_2222_socket(self):
        socket = start_socket("2222.us", 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.emit("test event", ["data0", "data1", 2])
        socket.emit("test event", ("data0", "data1", 2))
        socket.disconnect()

    @pytest.mark.skip("Server is decommissioned")
    def test_3333_socket(self):
        socket = start_socket("3333.us", 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.emit("test event", ["data0", "data1", 2])
        socket.emit("test event", ("data0", "data1", 2))
        socket.disconnect()

    @pytest.mark.skip("Server is decommissioned")
    def test_5555_socket(self):
        socket = start_socket("5555.us", 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.emit("test event", ["data0", "data1", 2])
        socket.emit("test event", ("data0", "data1", 2))
        socket.disconnect()

    def test_0000_socket(self):
        socket = start_socket("0000.us", 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.emit("test event", ["data0", "data1", 2])
        socket.emit("test event", ("data0", "data1", 2))
        socket.disconnect()

    @pytest.mark.skip("Server is decommissioned")
    def test_6666_socket(self):
        socket = start_socket("6666.us", 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.emit("test event", ["data0", "data1", 2])
        socket.emit("test event", ("data0", "data1", 2))
        socket.disconnect()


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import time
from typing import Optional

from neon_mq_connector.utils.rabbit_utils import create_mq_callback
from pika.exchange_type import ExchangeType
from neon_utils import LOG
from neon_mq_connector import MQConnector

from klat_connector.api.klat_api_abc import KlatApiABC


class KlatAPIMQ(KlatApiABC, MQConnector):
    def __init__(self, config: Optional[dict], service_name: str, vhost: str):
        """
        Create an API connection to Klat via RabbitMQ
        :param config: MQ Config dict
            ``` JSON Template of configuration:

                     { "users": {"<service_name>": { "user": "<username>",
                                                     "password": "<password>" }
                                }
                       "server": "localhost",
                       "port": 5672
                     }
            ```
        :param service_name: name of service credentials to use
        :param vhost: RabbitMQ vhost to connect to
        """
        # TODO: Spec default vhost to connect to
        MQConnector.__init__(self, config, service_name)
        self.current_conversations = set()
        self.vhost = vhost
        self._connected = False

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def nick(self) -> str:
        # TODO: Support login to arbitrary Klat account
        return self.uid

    @property
    def uid(self) -> str:
        return self.service_name + '-' + self.service_id

    @property
    def connection_exchange(self) -> str:
        return 'connection'

    @property
    def disconnection_exchange(self) -> str:
        return 'disconnection'

    def handle_incoming_shout(self, message_data: dict):
        """Handles incoming shout for this user"""
        LOG.info(f'Received message data: {message_data}')

    @create_mq_callback()
    def _on_user_message(self, body: dict):
        if body.get('cid', None) in self.current_conversations and (body.get('is_broadcast', False)
                                                                    or body.get('receiver', None) == self.nick):
            self.handle_incoming_shout(body)

    def send_shout(self, shout: str, cid: str = None, dom: str = None,
                   queue_name: str = 'user_message',
                   exchange: Optional[str] = None,
                   broadcast: bool = False, context: dict = None,
                   **kwargs):
        """
        Shout into the current conversation or else passed dom/cid
        :param shout: text to shout
        :param cid: CID to send shout into (Private cid for @user shout)
        :param dom: Domain associated with cid
        :param queue_name: name of the response mq queue
        :param exchange: name of mq exchange
        :param broadcast: to broadcast shout (defaults to False)
        :param context: message context to pass along with response
        :param kwargs: additional params to include in message body
            `bot_type`: Optional string BotTypes for chatbots
            `conversation_state`: Optional string ConversationState for cid
            `responded_shout`: Optional string messageID to reply to
        """
        # TODO: Handle current cid or most recent cid for dom as default?
        if not cid:
            raise ValueError("cid not specified")

        if broadcast:
            # prohibits fanouts to default exchange for consistency
            exchange = exchange if exchange else queue_name  # handle ''
            queue_name = ''
            exchange_type = ExchangeType.fanout.value
        else:
            exchange_type = ExchangeType.direct.value

        message_body = {
            'nick': self.uid,
            'service_name': self.service_name,
            'cid': cid,
            'dom': dom,
            'shout': shout,
            'context': context or dict(),
            'time': str(time.time()),
            'is_broadcast': broadcast,
            **kwargs
        }

        self._send_shout(queue_name=queue_name,
                         exchange=exchange,
                         exchange_type=exchange_type,
                         message_body=message_body)

    def _send_shout(self, queue_name: str = '', message_body: dict = None,
                    exchange: str = '',
                    exchange_type: str = ExchangeType.direct.value) -> str:
        """
        Sends shout from current instance

        :param queue_name: MQ queue name for emit (optional for publish=True)
        :param message_body: dict with relevant message data
        :param exchange: MQ exchange name for emit
        :param exchange_type: type of exchange to use based on ExchangeType
            (direct, fanout, topic...)
        :returns generated shout id
        """
        if not message_body:
            LOG.warning("Cannot send shout without message")
            return
        return self.send_message(
            request_data=message_body,
            vhost=self.vhost,
            exchange=exchange,
            queue=queue_name,
            exchange_type=exchange_type,
        )

    def run(self, *args, **kwargs):
        self._setup_listeners()
        MQConnector.run(self, *args, **kwargs)
        self._on_connect()

    def _start_connection(self):
        self.run(run_sync=False)

    def _stop_connection(self):
        self._on_disconnect()

    def _on_connect(self):
        self._send_shout(
            message_body={
                'nick': self.uid,
                'service_name': self.service_name,
                'time': int(time.time())
            },
            exchange=self.connection_exchange,
            exchange_type=ExchangeType.fanout,
        )
        self._connected = True

    def _on_disconnect(self):
        self._send_shout(
            message_body={
                'nick': self.uid,
                'service_name': self.service_name,
                'time': int(time.time())
            },
            exchange=self.disconnection_exchange,
            exchange_type=ExchangeType.fanout,
        )
        self._connected = False

    def _on_reconnect(self):
        self._stop_connection()
        self._start_connection()

    def _setup_listeners(self):
        self.register_consumer('user message', self.vhost, 'user_message',
                               self._on_user_message,
                               self.default_error_handler)
        self.register_consumer('user message broadcast', self.vhost, '',
                               self._on_user_message,
                               self.default_error_handler,
                               exchange="user_message",
                               exchange_type=ExchangeType.fanout.value)

    def shutdown(self):
        """
        Shutdown this connection; stop all consumers and close all connections
        """
        self.stop()

    def stop(self):
        self._stop_connection()
        MQConnector.stop(self)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from abc import ABC, abstractmethod
from typing import Optional, Union
from enum import IntEnum

from klat_connector.logger import LOG


class LoginState(IntEnum):
    DISCONNECTED = 0
    GUEST = 1
    USER = 2
    UNKNOWN = 3


class KlatApiABC(ABC):
    """Abstract class declaring the basic properties any inherited API should implement"""

    @property
    @abstractmethod
    def connected(self) -> bool:
        """
        Checks if instance is connected
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def nick(self) -> str:
        """
        Unique nickname for API instance in network
        """
        raise NotImplementedError

    @property
    def uid(self) -> str:
        """
        Override to specify a UID other than the login nick
        """
        return self.nick

    # Current Conversation properties
    @property
    def chat_log(self) -> Optional[list]:
        """
        Override to return a list of chat messages
        """
        return None

    @property
    def title(self) -> Optional[str]:
        """
        Override to return a conversation title
        """
        return None

    @property
    def logged_in(self) -> Union[int, LoginState]:
        """
        Override to return the login status of this instance
        """
        return LoginState.UNKNOWN

    @property
    def conversation_is_proctored(self) -> bool:
        """
        Override to return True if the conversation is proctored
        """
        return False

    @property
    def conversation_users(self) -> list:
        """
        Override to return a list of users in the current conversation
        """
        return []

    # Handlers to be overridden
    @abstractmethod
    def handle_incoming_shout(self, *args, **kwargs):
        """
        This function should be overridden to handle incoming messages
        """
        pass

    @abstractmethod
    def send_shout(self, shout: str, cid: str = None, dom: str = None,
                   **kwargs):
        """
        This method should be overridden to handle sending messages
        :param shout: text to shout
        :param cid: CID to send shout into (Private cid for @user shout)
        :param dom: Domain associated with cid
        """
        pass

    # Socket Listeners
    def _on_connect(self):
        """
        Handler for socket connection
        """
        LOG.info("Chat Server Socket Connected!")

    @staticmethod
    def _on_disconnect():
        """
        Handler for socket disconnection
        """
        # self.connected = False
        LOG.warning("Chat Server Socket Disconnected!")

    @staticmethod
    def _on_reconnect():
        """
        Handler for socket reconnection
        """
        # self.connected = True
        LOG.warning("SocketIO Reconnected")

    @abstractmethod
    def _on_user_message(self, *args):
        """
        Handler for "user message" (incoming shouts)
        :param args: Socket Arguments
        """
        pass

    @abstractmethod
    def _send_shout(self, *args, **kwargs):
        """
            Internal function that sends a shout into the conversation
        """
        pass

    @abstractmethod
    def _start_connection(self):
        """
        Initializes a new connection to the Klat server
        """
        pass

    @abstractmethod
    def _stop_connection(self):
        """
        Closes an open connection to the Klat server
        """
        pass

    @abstractmethod
    def _setup_listeners(self):
        """
        Starts all Klat event listeners
        """
        pass

    @abstractmethod
    def shutdown(self):
        """
        Clean up any pending transactions and close all connections.
        """

    @abstractmethod
    def run(self, *args, **kwargs):
        """
        Start any connections and perform any initialization that depends on
        connections. This may be called from `__init__`, or manually after
        object initialization.
        """



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from multiprocessing import Process
from time import time
from socketio import Server, WSGIApp
from eventlet import wsgi, listen
from ovos_utils.log import LOG


class MachKlatServer:
    def __init__(self, port: int = 8888):
        self.server = self._create_local_server()

        self._socket_users = dict()
        self._domain_to_cid = dict()
        self._p_title_to_cid = dict()
        self._cid_nicks = dict()

        self._set_server_event_listeners()
        self.t = Process(target=self._start_server_listener,
                         args=(self.server, port), daemon=True)
        self.t.start()
        LOG.debug("Mock Server Ready")

    def shutdown_server(self):
        """
        Terminates the server thread to end this process
        """
        self.t.terminate()

    @staticmethod
    def _start_server_listener(sio_server: Server, port: int):
        wsgi.server(listen(('0.0.0.0', port)), WSGIApp(sio_server))

    @staticmethod
    def _create_local_server() -> Server:
        """
        Creates a socketIO server
        :return: socketIO Server
        """
        return Server(async_mode="eventlet")

    @staticmethod
    def _generate_unique_nick() -> str:
        """
        Generates a unique nickname for a user
        :return: Random generated username
        """
        return f"guest{round(time())}"

    def _get_domain_cid(self, dom: str, title: str = None) -> str:
        """
        Get a cid associated with a given domain
        :param dom: domain requested
        :param title: optional title for Private domain use
        :return: cid
        """
        if dom == "Private":
            if title not in self._p_title_to_cid.keys():
                cid = str(round(time() * 1000))[5:]
                self._p_title_to_cid[title] = cid
            return self._p_title_to_cid.get(title)
        else:
            if not self._domain_to_cid.get(dom, None):
                self._domain_to_cid[dom] = str(round(time() * 1000))[5:]
            return self._domain_to_cid.get(dom)

    def _change_socket_nick(self, old_nick: str, new_nick: str, cid: str, sid: str):
        """
        Update internal references when a socket nick changes
        :param old_nick: old nick associated with socket
        :param new_nick: new nick associated with socket
        :param cid: cid socket is connected to
        :param sid: socketID associated with request
        :return:
        """
        new_nick = new_nick.lower()
        self._socket_users[sid] = new_nick
        self.server.enter_room(sid, new_nick)
        if not self._cid_nicks.get(cid):
            self._cid_nicks[cid] = list()
        if old_nick in self._cid_nicks[cid]:
            self._cid_nicks[cid].remove(old_nick)
        self._cid_nicks[cid].append(new_nick)

    def _get_nicks_for_cid(self, sid, *data):
        """
        Handle a request for nicks in the current cid
        :param sid: SocketID associated with request
        :param data: (cid requested)
        """
        # LOG.debug(f"{sid} get nicks")
        try:
            users = self._cid_nicks.get(data[0])
            LOG.debug(f"Got users: {users}")
            self.server.emit("new nick list", (None, users), sid)
        except Exception as e:
            LOG.error(e)

    def _check_login(self, sid, *data):
        """
        Handle a login request
        :param sid: SocketID associated with request
        :param data: (curr_nick, login_nick, curr_dom, hash_pass, ?, curr_cid, dev_uid)
        """
        LOG.info(f"{data[0]} Check Login as {data[1]}")
        old_nick = data[0]
        nick = data[1]
        cid = data[5]
        self._change_socket_nick(old_nick, nick, cid, sid)
        self.server.emit("mobile login return", ("mock", nick), sid)

    def _logout(self, sid, *data):
        """
        Handle a logout request
        :param sid: SocketID associated with request
        :param data: (nick, domain, ?, cid, instanceID, col#, #col)
        """
        LOG.info(f"{sid} Logout")
        cid = data[3]
        old_nick = data[0]
        new_nick = self._generate_unique_nick()
        self._change_socket_nick(old_nick, new_nick, cid, sid)
        self.server.emit("log out successful", (), sid)

    def _get_conversation_list(self, sid, *data):
        """
        Handle a request for conversation list
        :param sid: SocketID associated with request
        :param data: (pCid, cookieCid, mode, dom, nick, tab, hash, url, klatGuid, pageName, column, kinstance,
            allCookies, nano, welcomeMessage, fn)
        """
        LOG.info(f"{sid} Conversation List")
        cid = self._get_domain_cid(data[3])
        self.server.enter_room(sid, cid)
        conversations = [{"row": {"dom": data[3], "title": "Mock Server Testing", "cid": cid}}]
        self.server.emit("conversation list", (conversations, None, sid), sid)

    def _nickname(self, sid, *data):
        """
        Handle a request for nickname
        :param sid: SocketID associated with request
        :param data: (myCNick, luname, ldom, myConversation, userTitle, myFollowsKey, myColumn, myKinstance, myPageName,
            myTtsLanguage, mySecondTtsLanguage, mySttLanguage, nano, uid, function)
        """
        LOG.debug(f"{sid} Nickname")
        LOG.debug(data)
        nick = self._generate_unique_nick()
        self._socket_users[sid] = nick
        LOG.info(self._socket_users)
        devices = None
        profile = {}
        self.server.emit("nickname", (nick, devices, profile), sid)

    def _user_message(self, sid, *data):
        """
        Handle an incoming user message
        :param sid: SocketID associated with request
        :param data: (msg, dom, cid, img, video, mode, imgData, showAnswers, audioBlobs, audioFileName, sttLanguage,
            ttsLanguage, ttsSecondLanguage, ttsGender, needsTranslation, cidForLink, cookies, name, fn)
        """
        LOG.debug(f"{data[2]} | {data[0]}")
        shout = data[0]
        user = self._socket_users.get(sid, "UNDEF")
        cid = str(data[2])
        dom = data[1]
        millis_time = round(time() * 1000)
        # for socket in self._socket_users.keys():
        #     if socket != sid:
        # for socket in self._socket_users.keys():
        #     LOG.debug(f"{self._socket_users[socket]} is in: {self.server.rooms(socket)}")
        if shout.startswith("@"):
            username = shout.split(" ", 1)[0].lstrip("@").lower()
            LOG.info(f"Sending shout to room: {username}")

            for sid in self._socket_users.keys():
                rooms = self.server.rooms(sid)
                if username in rooms:
                    LOG.info(f"DM to {username} | {sid}")
            self.server.emit("mobile user message", (shout, user, cid, millis_time, None, dom), room=username)
        else:
            self.server.emit("mobile user message", (shout, user, cid, millis_time, None, dom), room=cid)

    def _create_conversation(self, sid, *data):
        """
        Handle a request to create a conversation
        :param sid: SocketID associated with request
        :param data: (domain, nick, cid, title, ImageURL, ArticleURL, msgFromNeon, PassHash, FirstShout, keychain)
        """
        LOG.info(f"Creating conversation {data}")
        new_cid = self._get_domain_cid(data[0], data[3])
        new_dom = data[0]
        self.server.emit("create conversation return", (new_cid, "title", None, new_dom), sid)

    def _change_domain(self, sid, *data):
        """
        Handle a request to change domains
        :param sid: SocketID associated with request
        :param data: ("this one", curr_cid, new_dom)
        """
        LOG.info(f"{sid} changing domains to {data[2]}")
        cid = self._get_domain_cid(data[2])
        dom = data[2]
        self.server.leave_room(sid, data[1])
        if self._socket_users[sid] in self._cid_nicks.get(cid, []):
            self._cid_nicks[cid].remove(self._socket_users[sid])
        self.server.enter_room(sid, cid)
        if not self._cid_nicks.get(cid):
            self._cid_nicks[cid] = list()
        self._cid_nicks[cid].append(self._socket_users[sid])

        if dom == "Private":
            title = f"!PRIVATE:{self._socket_users[sid]}"
        else:
            title = "Title"
        self.server.emit("new domain", (cid, None, {"dom": dom, "title": title}), sid)

    def _on_disconnect(self, sid, *data):
        """
        Handle a disconnected socket
        :param sid: SocketID disconnected
        :param data: ()
        """
        LOG.info(f"{sid} disconnected!")
        LOG.debug(data)
        try:
            self._socket_users.pop(sid)
        except Exception as e:
            LOG.error(e)

    def _search_shouts_by_phrase(self, sid, *data):
        """
        Handle a search request
        :param sid: SocketID associated with request
        :param data: (search_term, num_context, max_time)
        """
        LOG.debug(data)
        self.server.emit("search shouts by phrase return", ({}), sid)

    def _save_login(self, sid, *data):
        """
        Handle a Klat registration request
        :param sid: SocketID associated with request
        :param data: (username, domain, hash_pass, color, cid, kinstance, col1, 1col, 1?, 1?, speech_rate, stt_engine,
            tts_name, tts_gender, tts_language, tts_language_2, stt_language, time_format, units, date_format,
            city, state, country, uid, ?)
        """
        new_uname = data[0].lower()
        cid = data[4]
        old_nick = self._socket_users[sid]
        self._change_socket_nick(old_nick, new_uname, cid, sid)
        self.server.emit("save login return", (None, data[0], "Save login successful."), sid)

    def _get_log(self, sid, *data):
        """
        Handle a conversation log request
        :param sid: SocketID associated with request
        :param data: (myDomain, oldConversation, myConversation, myNick, myFollowsKey, myColumn,
            myMaxFollowedConversations, myContextMax, myContextTime, nano, lastShout, numShouts, function(err){})
        """
        dom = data[0]
        c_users = self._cid_nicks.get(dom)
        self.server.emit("chat log", ([], c_users), sid)

    def _set_server_event_listeners(self):
        self.server.on("get nicks for cid", self._get_nicks_for_cid)
        self.server.on("check login", self._check_login)
        self.server.on("logout", self._logout)
        self.server.on("nickname", self._nickname)
        self.server.on("user message", self._user_message)
        self.server.on("get conversation list 2", self._get_conversation_list)
        self.server.on("create conversation", self._create_conversation)
        self.server.on("change domain", self._change_domain)
        self.server.on("disconnect", self._on_disconnect)
        self.server.on("search shouts by phrase", self._search_shouts_by_phrase)
        self.server.on("save login", self._save_login)
        self.server.on("get log", self._get_log)


if __name__ == "__main__":
    server = MachKlatServer()
    server.t.join()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import hashlib
import struct

from typing import List
from ovos_utils.log import LOG


def generate_hash(to_be_hashed: str) -> List[int]:
    """
    Generates an SHA hash for the input string
    :param to_be_hashed: string to be hashed
    :return: list(int) hash
    """
    md = hashlib.sha512()
    input_array = bytearray()
    input_array.extend(map(ord, to_be_hashed))
    LOG.debug(list(input_array))
    md.update(bytearray(input_array))
    byte_data = md.digest()
    num_ints = int(len(byte_data))
    hashed = list(struct.unpack("b"*num_ints, byte_data))
    LOG.debug(hashed)
    return hashed


def encrypt(string):
    # TODO: Implement encrypt method
    return string


def decrypt(string):
    # TODO: Implement decrypt method
    return string



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import unittest
import os
from threading import Event

import psutil
import pytest

from multiprocessing import Process
from time import time
from socketio import Client
from unittest.mock import Mock
from ovos_utils.log import LOG
from neon_minerva.integration.rabbit_mq import rmq_instance

from klat_connector.utils.sio_connection_utils import start_socket
from klat_connector.api.sio_klat_api import KlatApi
from klat_connector.api.mq_klat_api import KlatAPIMQ
from klat_connector.utils.legacy_socket_adapter import SocketIOCompat

SERVER = os.environ.get("server") or "2222.us"
os.environ["TEST_RMQ_USERNAME"] = "test_user"
os.environ["TEST_RMQ_PASSWORD"] = "test_password"
os.environ["TEST_RMQ_VHOSTS"] = "/test_chatbots"


class KlatApiTest(KlatApi):
    def __init__(self, *args, **kwargs):
        self.mock_method = kwargs.pop("mock")
        super().__init__(*args, **kwargs)

    def handle_incoming_shout(self, user: str, shout: str, cid: str, dom: str, timestamp: str):
        LOG.info(f"{user}|{shout}|{cid}|{dom}|{timestamp}")
        self.mock_method(shout, user, dom, cid)


@pytest.mark.skip(reason="Klat 1.0 Test server has been decommissioned")
class SIOKlatAPITests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Obfuscate server address for GHA secret filtering
        LOG.info(f"server={'|'.join(list(SERVER))}")
        cls.socket = start_socket(SERVER)
        cls.klat_api = KlatApi(cls.socket, "Private")

    @classmethod
    def tearDownClass(cls) -> None:
        cls.socket.disconnect()

    def setUp(self) -> None:
        if not self.socket.connected:
            LOG.warning("Socket not connected, recreating")
            self.socket = start_socket(SERVER)
            self.klat_api = KlatApi(self.socket, "Private")

    # @pytest.mark.timeout(30)
    # def test_start_socket(self):
    #     socket = start_socket()
    #     self.assertIsInstance(socket, Client)
    #     self.assertTrue(socket.connected)
    #     socket.disconnect()

    def test_start_socket_with_args(self):
        socket = start_socket(SERVER, 8888)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertTrue(socket.connected)
        socket.disconnect()

    def test_start_socket_bad_port_timeout(self):
        socket = start_socket(SERVER, 80, 5)
        self.assertIsInstance(socket, (Client, SocketIOCompat))
        self.assertFalse(socket.connected)

    def test_00_klat_api_properties(self):
        self.assertTrue(self.klat_api.connected)
        self.assertIsNone(self.klat_api.chat_log)
        self.assertIsInstance(self.klat_api.nick, str)
        self.assertTrue(self.klat_api.nick.startswith("guest"))
        self.assertIsInstance(self.klat_api.title, str)
        self.assertTrue(self.klat_api.title.startswith("!PRIVATE"))
        self.assertEqual(self.klat_api.logged_in, 1)
        self.assertFalse(self.klat_api.conversation_is_proctored)
        self.assertIsInstance(self.klat_api.conversation_users, list)
        self.assertIn(self.klat_api.nick, self.klat_api.conversation_users)

    def test_klat_api_connection_log(self):
        self.assertTrue(self.socket.connected)

        klat = self.klat_api
        klat.change_domain("Private")
        self.assertEqual(klat.logged_in, 1, "Error Connecting as Guest")
        self.assertIsInstance(klat.conversation_users, list, "Users not Populated")
        self.assertIsInstance(klat.nick, str, "Invalid Nick")
        self.assertFalse(klat.conversation_is_proctored, "Invalid Private Proctored conversation")
        self.assertFalse(klat.is_current_cid(""))

        klat.get_conversation_log()
        self.assertIsInstance(klat.chat_log, list, "Chat Log Invalid!")

    def test_klat_login(self):
        klat = self.klat_api
        self.assertTrue(klat.socket.connected)

        klat.login_klat("testrunner", "testpassword")
        self.assertEqual(klat.logged_in, 2, "Error logging in")
        self.assertTrue(klat.socket.connected)
        klat.logout_klat()
        self.assertEqual(klat.logged_in, 1, "Error logging out")
        self.assertTrue(klat.nick.startswith("guest"), "Not a guest after logging out")

        nick = klat.nick
        klat.login_klat("testrunner", "incorrectPassword")
        self.assertEqual(klat.logged_in, 1, "Unknown exception with wrong password")
        self.assertEqual(nick, klat.nick, "Nick error after failed login!")
        klat.logout_klat()

    def test_klat_register(self):
        # TODO: Troubleshoot this with shared socket/api instance DM
        klat = KlatApi(start_socket(SERVER))
        username = f"testrunner{time()}".split(".")[0]
        self.assertTrue(klat.socket.connected)
        self.assertIsNotNone(klat._dom)
        self.assertTrue(klat.nick.startswith("guest"))

        klat.register_klat(username, "testpassword")
        self.assertEqual(klat.logged_in, 2)
        self.assertEqual(klat.nick, username)
        klat.logout_klat()
        self.assertEqual(klat.logged_in, 1, "Error logging out")
        self.assertTrue(klat.nick.startswith("guest"), "Not a guest after logging out")

        nick = klat.nick
        klat.register_klat(username, "testpassword")
        self.assertEqual(klat.logged_in, 1, "Unknown exception with wrong password")
        self.assertEqual(nick, klat.nick, "Nick error after failed login!")
        klat.socket.disconnect()

    def test_change_conversation(self):
        klat = self.klat_api
        klat.change_domain("Private")
        cid = klat._cid
        self.assertEqual(klat._dom, "Private")
        klat.change_conversation("47562", "klattalk.com")
        self.assertEqual(klat._dom, "klattalk.com")
        self.assertEqual(klat._cid, "47562")

        klat.change_conversation(cid, "Private")
        self.assertEqual(klat._cid, cid)

    def test_change_domain(self):
        klat = self.klat_api
        klat.change_domain("CatTalk.com")
        self.assertEqual(klat._dom, "cattalk.com")
        klat.change_domain("private")
        self.assertEqual(klat._dom, "Private")

    def test_create_goto_conversation(self):
        klat = self.klat_api
        klat.change_domain("klattalk.com")
        new_name = f'NewConvoTest {str(time()).split(".")[0]}'
        cid = klat.create_conversation("klattalk.com", new_name, True)
        self.assertIsNotNone(cid)
        self.assertTrue(klat.is_current_cid(cid))

    def test_create_invalid_conversation(self):
        klat = self.klat_api
        klat.change_domain("klattalk.com")
        cid = klat.create_conversation("Private", "NewConvoTest", True)
        self.assertIsNone(cid)

    def test_create_private_conversation(self):
        klat = self.klat_api
        cid = klat.get_private_conversation(["testrunner"])
        self.assertIsNotNone(cid)
        klat.change_conversation(cid, "Private")

        self.assertEqual(klat._dom, "Private")
        self.assertEqual(klat._cid, cid)

    def test_is_current_cid(self):
        klat = self.klat_api
        cid = klat._cid
        is_current = klat.is_current_cid(cid)
        self.assertTrue(is_current)
        isnt_current = klat.is_current_cid("1")
        self.assertFalse(isnt_current)

    def test_conversation_is_proctored(self):
        klat = self.klat_api
        klat.change_domain("Private")
        self.assertFalse(klat.conversation_is_proctored)

    def test_conversation_users(self):
        klat = self.klat_api
        klat.change_domain("Private")
        users = klat.conversation_users
        self.assertEqual(users, [klat.nick], f"cid={klat._cid}|title={klat.title}")
        klat.change_domain("chatbotsforum.org")
        self.assertIsInstance(klat.conversation_users, list)
        self.assertTrue(len(klat.conversation_users) > 0)
        self.assertIn(klat.nick, klat.conversation_users)

    @pytest.mark.timeout(15)
    def test_send_shout(self):
        test_dom = "aatalk.com"
        test_str = "Hello!"
        mock_method = Mock()
        called = Event()
        mock_method.side_effect = lambda: called.set()

        klat_api = KlatApiTest(start_socket(SERVER), test_dom, mock=mock_method)
        self.klat_api.change_domain(test_dom)
        self.klat_api.send_shout(test_str)
        called.wait(10)
        klat_api.socket.disconnect()
        mock_method.assert_called_once()
        mock_method.assert_called_with(test_str, self.klat_api.nick, test_dom, self.klat_api._cid)

    def test_send_shout_disconnected(self):
        klat = KlatApi(start_socket(SERVER), "Private")
        klat.socket.disconnect()
        with pytest.raises(ConnectionError) as e:
            klat._send_shout("No Connection!", klat._dom, klat._cid, klat._cid, None)
            self.assertTrue(str(e).startswith("Socket disconnected skipping sending:"))

    @pytest.mark.xfail
    @pytest.mark.timeout(30)
    def test_multi_socket_connections(self):
        def run_test():
            klat = KlatApi(start_socket(SERVER), "Private")
            cid = klat._cid
            self.assertEqual(klat._dom, "Private")
            self.assertEqual(klat.conversation_users, [klat.nick], klat._cid)
            klat.change_domain("Private")
            self.assertEqual(klat._cid, cid)
            self.assertEqual(klat.title.rsplit('-', 1)[0], f"!PRIVATE:{klat.nick}")
            klat.socket.disconnect()
            return

        processes = []
        for i in range(2):
            p = Process(target=run_test)
            p.start()
            processes.append(p)

        success = True
        for p in processes:
            p.join(10)
            if p.is_alive():
                psutil.Process(p.pid).kill()
                success = False
        self.assertTrue(success)


@pytest.mark.usefixtures("rmq_instance")
class MQKlatAPITests(unittest.TestCase):
    klat_api = None
    service_name = "chatbot"
    vhost = "/test_chatbots"
    mock_handled = Event()
    mq_config = {"users": {"chatbot": {"user": "test_user",
                                       "password": "test_password"}},
                 "server": "127.0.0.1"
                 }

    @classmethod
    def tearDownClass(cls) -> None:
        cls.klat_api.shutdown()

    def setUp(self) -> None:
        if MQKlatAPITests.klat_api is None:
            self.mq_config["port"] = self.rmq_instance.port

            MQKlatAPITests.klat_api = KlatAPIMQ(self.mq_config,
                                                self.service_name, self.vhost)
            MQKlatAPITests.klat_api.handle_incoming_shout = Mock(
                side_effect=self.mock_handled.set)
            MQKlatAPITests.klat_api.run()

        MQKlatAPITests.klat_api.handle_incoming_shout.reset_mock()

    def test_00_init(self):
        self.assertTrue(self.klat_api.connected)
        self.assertEqual(self.klat_api.current_conversations, set())
        self.assertIsInstance(self.klat_api.nick, str)
        self.assertIsInstance(self.klat_api.uid, str)
        self.assertIn("user message", self.klat_api.consumers)

    def test_messaging(self):
        shout = "Test message.\n"
        cid = "test_cid"
        self.klat_api.current_conversations.add(cid)

        def _validate_shout_data(data: dict):
            self.assertEqual(data["nick"], self.klat_api.uid)
            self.assertEqual(data["service_name"], self.klat_api.service_name)
            self.assertEqual(data["cid"], cid)
            self.assertEqual(data["shout"], shout)
            self.assertEqual(data["context"], dict())
            self.assertIsInstance(data["time"], str)
            self.assertIsNone(data["dom"])

        # Direct Message to other user
        self.mock_handled.clear()
        self.klat_api.send_shout(shout, cid, receiver="other_user_id")
        self.mock_handled.wait(5)
        self.klat_api.handle_incoming_shout.assert_not_called()

        # Direct Message
        self.mock_handled.clear()
        self.klat_api.send_shout(shout, cid, receiver=self.klat_api.uid)
        self.mock_handled.wait(5)
        self.klat_api.handle_incoming_shout.assert_called_once()
        call_args = self.klat_api.handle_incoming_shout.call_args[0][0]
        _validate_shout_data(call_args)
        self.assertEqual(call_args["receiver"], self.klat_api.uid)
        self.assertFalse(call_args["is_broadcast"])

        # Fanout Message
        self.klat_api.handle_incoming_shout.reset_mock()
        self.mock_handled.clear()
        self.klat_api.send_shout(shout, cid, broadcast=True)
        self.mock_handled.wait(5)
        self.klat_api.handle_incoming_shout.assert_called_once()
        call_args = self.klat_api.handle_incoming_shout.call_args[0][0]
        _validate_shout_data(call_args)
        self.assertIsNone(call_args.get("receiver"))
        self.assertTrue(call_args["is_broadcast"])


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from threading import Thread
from socketIO_client import SocketIO


class SocketIOCompat(SocketIO):
    def __init__(self, *args, **kwargs):
        super(SocketIOCompat, self).__init__(*args, **kwargs)

    def emit(self, event, data, namespace=None, callback=None):
        """
        Adapts the new syntax expected by socketio.client to the legacy socketio_client package
        """
        if isinstance(data, list) or isinstance(data, tuple):
            super().emit(event, *data, path=namespace, callback=callback)
        elif isinstance(data, dict):
            super().emit(event, **data, path=namespace, callback=callback)
        else:
            super().emit(event, data, path=namespace, callback=callback)

    def send(self, data, namespace=None, callback=None):
        super().send(data, callback, path=namespace)

    def connect(self, url):
        super().connect()

    def _run_listener(self):
        self.wait()

    def run_forever(self):
        event_thread = Thread(target=self._run_listener)
        event_thread.daemon = True
        event_thread.start()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2025 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from socketio import exceptions as sio_exceptions
from typing import Union
from socketio import Client
from threading import Thread
from time import time, sleep
from ovos_utils.log import LOG
from klat_connector.utils.legacy_socket_adapter import SocketIOCompat


# Default server connection
_server_addr = "0000.us"
_server_port = 8888


def _run_listener(socket):
    """
    Runs socket listeners indefinitely
    """
    socket.wait()


def _establish_socket_connection(socket: Client, addr: str, port: int):
    """
    Starts a thread to connect and run the socket connection infinitely.
    :param socket: SocketIO client object to connect
    :param addr: IP, hostname, or URL of SocketIO server
    :param port: host port to connect to
    """
    if addr == "0.0.0.0":
        url = f"http://{addr}:{port}"
    else:
        url = f"https://{addr}:{port}"
    socket.connect(url, wait_timeout=10, transports=['websocket', 'polling'],
                   namespaces='/')
    event_thread = Thread(target=_run_listener, args=[socket])
    event_thread.daemon = True
    event_thread.start()
    LOG.debug(f"returning {socket}")
    return socket


def _establish_legacy_socket_connection(socket: SocketIOCompat):
    """
    Establish a legacy SocketIO connection
    :param socket: SocketIOCompat object to connect
    """
    socket.connect(None)


def start_socket(addr=_server_addr, port=_server_port, retry_timeout=120) -> \
        Union[Client, SocketIOCompat]:
    """
    Initialize a socketIO connection to the specified server and port
    :param addr: url of socketIO server to connect to
    :param port: port used for socketIO
    :param retry_timeout: max seconds to try to establish a connection
    :return: socketIO Client
    """
    # global socket
    socket = Client(request_timeout=10)

    timeout = time() + retry_timeout

    # Catch connected socket
    while not socket.connected and time() < timeout:
        try:
            if isinstance(socket, Client):
                _establish_socket_connection(socket, addr, port)
            elif isinstance(socket, SocketIOCompat):
                _establish_legacy_socket_connection(socket)
        except sio_exceptions.ConnectionError as e:
            if isinstance(socket, Client) and e.args[0] == "Unexpected response from server":
                LOG.warning("Falling back to SocketIOCompat connection")
                socket = SocketIOCompat(f"https://{addr}", port, verify=False)
            else:
                LOG.exception(e)
                LOG.error("Retrying in 5 seconds")
                sleep(5)
    return socket
